import sys
from io import StringIO
from pathlib import Path
from unittest import TestCase

from mypy.api import run as mypy_api_run


class TestMypyRaisePlugin(TestCase):
    def run_mypy(self, resource_file: str) -> str:
        """Run mypy on a resource file using default config and return stdout."""
        return self.run_mypy_with_config(resource_file, 'mypy.ini')

    def run_mypy_with_config(self, resource_file: str, config_file: str) -> str:
        """Run mypy on a resource file with a custom config file and return stdout."""
        resource_path = Path(__file__).resolve().parent / 'resources' / resource_file
        config_path = Path(__file__).resolve().parent / 'resources' / config_file

        capture = StringIO()
        old_stdout = sys.stdout
        sys.stdout = capture

        try:
            stdout, stderr, exit_status = mypy_api_run(
                [
                    '--config-file',
                    str(config_path),
                    '--no-error-summary',
                    '--hide-error-context',
                    '--no-incremental',
                    str(resource_path),
                ]
            )
            combined_stdout = stdout + capture.getvalue()
            return combined_stdout
        finally:
            sys.stdout = old_stdout

    def test_safe_function(self):
        output = self.run_mypy('safe_function.py')
        # mypy.api returns empty string when there are no issues
        self.assertEqual(output.strip(), '')

    def test_explicit_raise_in_safe_function(self):
        output = self.run_mypy('explicit_raise_in_safe_function.py')
        self.assertIn('may raise', output)
        self.assertIn('not declared', output)

    def test_call_unsafe_function(self):
        output = self.run_mypy('call_unsafe_function.py')
        # Since 'unknown' is not decorated, we can't track its exceptions
        # The current implementation doesn't report this as an error
        # because we only track decorated functions
        self.assertEqual(output.strip(), '')

    def test_call_safe_function(self):
        output = self.run_mypy('call_safe_function.py')
        self.assertEqual(output.strip(), '')

    def test_call_explicit_raiser(self):
        output = self.run_mypy('call_explicit_raiser.py')
        self.assertIn('may raise', output)
        self.assertIn('not declared', output)

    def test_no_raise_calls_raiser(self):
        """Test function with no exceptions calling a raiser."""
        output = self.run_mypy('no_raise_calls_raiser.py')
        self.assertIn('may raise', output)
        self.assertIn('not declared', output)

    def test_indirect_exception_propagation(self):
        """Test exception propagation through multiple call levels."""
        output = self.run_mypy('indirect_exception_propagation.py')
        self.assertIn('may raise', output)
        self.assertIn('not declared', output)

    def test_stdlib_exception(self):
        """Test calling stdlib function that raises exceptions."""
        output = self.run_mypy('stdlib_exception.py')
        # Should detect exceptions from open()
        self.assertIn('may raise', output)
        self.assertIn('not declared', output)

    def test_correct_exception_declaration(self):
        """Test function correctly declaring all exceptions."""
        output = self.run_mypy('correct_exception_declaration.py')
        self.assertEqual(output.strip(), '')

    def test_missing_exception_declaration(self):
        """Test function missing exception in declaration."""
        output = self.run_mypy('missing_exception_declaration.py')
        self.assertIn('may raise', output)
        self.assertIn('not declared', output)

    def test_bare_raise(self):
        """Test bare raise statement handling."""
        output = self.run_mypy('bare_raise.py')
        # Bare raise should be tracked as 'Unknown' exception
        self.assertIn('may raise', output)

    def test_chained_attributes(self):
        """Test chained attribute resolution (os.path.join)."""
        output = self.run_mypy('chained_attributes.py')
        # os.path.join doesn't raise exceptions
        self.assertEqual(output.strip(), '')

    def test_module_exception(self):
        """Test module.Exception format exception name."""
        output = self.run_mypy('module_exception.py')
        # Should correctly parse builtins.ValueError
        self.assertEqual(output.strip(), '')

    def test_none_tracking(self):
        """Test exceptions=None case (not tracked)."""
        output = self.run_mypy('none_tracking.py')
        # Functions with exceptions=None should not be tracked
        self.assertEqual(output.strip(), '')

    def test_async_functions(self):
        """Test @raising decorator on async functions."""
        output = self.run_mypy('async_functions.py')
        # Should detect ValueError from async_method call
        self.assertIn('may raise', output)
        self.assertIn('ValueError', output)

    def test_property_methods(self):
        """Test @raising decorator on property methods."""
        output = self.run_mypy('property_methods.py')
        # Should detect ValueError from property access
        self.assertIn('may raise', output)
        self.assertIn('ValueError', output)

    def test_classmethod_functions(self):
        """Test @raising decorator on class methods."""
        output = self.run_mypy('classmethod_functions.py')
        # Should detect ValueError from classmethod call
        self.assertIn('may raise', output)
        self.assertIn('ValueError', output)

    def test_staticmethod_functions(self):
        """Test @raising decorator on static methods."""
        output = self.run_mypy('staticmethod_functions.py')
        # Should detect TypeError from staticmethod call
        self.assertIn('may raise', output)
        self.assertIn('TypeError', output)

    def test_instance_methods(self):
        """Test @raising decorator on instance methods."""
        output = self.run_mypy('instance_methods.py')
        # Should detect ValueError and TypeError from instance method calls
        self.assertIn('may raise', output)
        self.assertIn('ValueError', output)

    def test_custom_config(self):
        """Test custom exception configuration via mypy.ini."""
        output = self.run_mypy_with_config('custom_config.py', 'test_mypy.ini')
        # Should detect CustomError and ValueError from third_party_function (configured in mypy.ini)
        self.assertIn('may raise', output)
        self.assertIn('CustomError', output)
        self.assertIn('ValueError', output)

    def test_exception_caught(self):
        """Test that exceptions caught in try-except don't propagate."""
        output = self.run_mypy('exception_caught.py')
        # Functions without @raising decorator are not tracked
        # So there should be no errors even though risky_function raises ValueError
        self.assertEqual(output.strip(), '')

    def test_try_except_handling(self) -> None:
        """Test handling of exceptions within try-except blocks."""
        output = self.run_mypy('try_except_handling.py')
        
        # Should detect TypeError not caught in multiple_exceptions_caught
        self.assertIn("Function 'multiple_exceptions_caught' may raise 'TypeError'", output)
        
        # Should detect incorrect exception caught in wrong_exception_caught
        self.assertIn("Function 'wrong_exception_caught' may raise 'ValueError'", output)
        
        # Should NOT report ValueError for handles_exception (it's caught)
        self.assertNotIn("Function 'handles_exception' may raise", output)
        
        # Should NOT report exceptions for tuple_exception_handler
        self.assertNotIn("Function 'tuple_exception_handler' may raise", output)

    def test_exception_hierarchy(self) -> None:
        """Test invalid exception hierarchy handling (current reproduction)."""
        output = self.run_mypy('exception_hierarchy.py')
        
        # These SHOULD PASS if hierarchy is working, but currently they FAIL
        # because we only check strict equality.
        
        # handles_value_error_with_base catches Exception, raises ValueError
        # Once fixed, this assertion should be assertNotIn
        self.assertNotIn("Function 'handles_value_error_with_base' may raise 'ValueError'", output)
        
        # handles_file_error_with_oserror catches OSError, raises FileNotFoundError
        self.assertNotIn("Function 'handles_file_error_with_oserror' may raise 'FileNotFoundError'", output)
        
        # handles_key_error_with_lookup catches LookupError, raises KeyError
        self.assertNotIn("Function 'handles_key_error_with_lookup' may raise 'KeyError'", output)

    def test_strict_mode(self):
        """Test strict mode requires @raising decorator."""
        output = self.run_mypy_with_config('strict_mode.py', 'strict_mypy.ini')
        self.assertIn("Function 'undecorated_function' missing @raising decorator", output)
        self.assertIn("Function 'another_undecorated' missing @raising decorator", output)

    def test_declaration_hierarchy(self) -> None:
        """Test declaration with base exception covers subclasses."""
        output = self.run_mypy('declaration_hierarchy.py')
        # Expectation: NO errors because catching Exception should cover ValueError
        # Currently fails (contains errors)
        self.assertNotIn("may raise", output)

    def test_multiline_config(self):
        """Test parsing of multiline known_exceptions config."""
        output = self.run_mypy_with_config('multiline_config.py', 'multiline_mypy.ini')
        # third_party_lib.process is configured to raise ValueError
        self.assertIn("Function 'use_multiline_config' may raise 'ValueError'", output)
        self.assertIn("Raised by:", output)
        self.assertIn("'third_party_lib.process' raises 'ValueError'", output)




from typing import TypeAlias

FuncName: TypeAlias = str
LineNo: TypeAlias = int
ExceptionName: TypeAlias = str
ExceptionSet: TypeAlias = set[ExceptionName]
RaisingFunctionsLineno: TypeAlias = dict[FuncName, LineNo]
FunctionExceptionsMap: TypeAlias = dict[FuncName, ExceptionSet]
CallGraph: TypeAlias = dict[FuncName, list[tuple[LineNo, FuncName]]]
ExplicitRaises: TypeAlias = dict[FuncName, list[tuple[LineNo, ExceptionName]]]
ExceptionPropagationMap: TypeAlias = dict[FuncName, ExceptionSet]



from dataclasses import dataclass


@dataclass
class AnalysisStats:
    """Statistics for exception analysis."""

    files_checked: int = 0
    functions_checked: int = 0
    violations_found: int = 0
    exceptions_caught: int = 0

    def compliance_rate(self) -> float:
        """Calculate compliance rate."""
        if self.functions_checked == 0:
            return 100.0
        compliant = self.functions_checked - self.violations_found
        return (compliant / self.functions_checked) * 100

    def format_summary(self) -> str:
        """Format statistics summary."""
        lines = [
            '\n' + '=' * 60,
            'mypy-raise Analysis Summary',
            '=' * 60,
            f'Files checked: {self.files_checked}',
            f'Functions analyzed: {self.functions_checked}',
            f'Violations found: {self.violations_found}',
            f'Compliance rate: {self.compliance_rate():.1f}%',
            '=' * 60,
        ]
        return '\n'.join(lines)


# Global statistics instance
STATS = AnalysisStats()



import ast

from mypy_raise.types import (
    CallGraph,
    ExceptionSet,
    ExplicitRaises,
    FuncName,
    FunctionExceptionsMap,
    LineNo,
)


class RaisingVisitor(ast.NodeVisitor):
    def __init__(self) -> None:
        # func_name -> set of declared exception names
        self.__function_exceptions: FunctionExceptionsMap = {}
        # func_name -> lineno where decorator is
        self.__function_linenos: dict[FuncName, LineNo] = {}
        # func_name -> list of (lineno, called_func_name)
        self.__calls: CallGraph = {}
        # func_name -> list of (lineno, exception_name) for explicit raises
        self.__explicit_raises: ExplicitRaises = {}
        # func_name -> set of exception names caught in try-except blocks
        self.__caught_exceptions: FunctionExceptionsMap = {}

        self.__current_function: FuncName | None = None

    @property
    def function_exceptions(self) -> FunctionExceptionsMap:
        """Map of function names to their declared exceptions."""
        return self.__function_exceptions

    @property
    def function_linenos(self) -> dict[FuncName, LineNo]:
        """Map of function names to their decorator line numbers."""
        return self.__function_linenos

    @property
    def calls(self) -> CallGraph:
        return self.__calls

    @property
    def explicit_raises(self) -> ExplicitRaises:
        return self.__explicit_raises

    @property
    def caught_exceptions(self) -> FunctionExceptionsMap:
        """Map of function names to exceptions they catch in try-except blocks."""
        return self.__caught_exceptions

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self._process_function_def(node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self._process_function_def(node)

    def _process_function_def(self, node: ast.FunctionDef | ast.AsyncFunctionDef) -> None:
        """Process both regular and async function definitions."""
        previous_function = self.__current_function
        self.__current_function = node.name

        # Check for @raising decorator
        for decorator in node.decorator_list:
            if isinstance(decorator, ast.Call):
                # Check if it's the @raising decorator
                if isinstance(decorator.func, ast.Name) and decorator.func.id == 'raising':
                    # Extract the exceptions parameter
                    exceptions_arg = None
                    for keyword in decorator.keywords:
                        if keyword.arg == 'exceptions':
                            exceptions_arg = keyword.value
                            break

                    if exceptions_arg is not None:
                        # Parse the exceptions list
                        if isinstance(exceptions_arg, ast.Constant) and exceptions_arg.value is None:
                            # exceptions=None means don't track this function
                            pass
                        elif isinstance(exceptions_arg, ast.List):
                            # Extract exception names from the list
                            exception_names: ExceptionSet = set()
                            for elt in exceptions_arg.elts:
                                exc_name = self.__extract_exception_name(elt)
                                if exc_name:
                                    exception_names.add(exc_name)

                            self.__function_exceptions[node.name] = exception_names
                            self.__function_linenos[node.name] = decorator.lineno

        # Only visit the body to avoid treating decorators as calls inside the function
        for item in node.body:
            self.visit(item)

        self.__current_function = previous_function

    def __extract_exception_name(self, node: ast.AST) -> str | None:
        """Extract exception name from AST node."""
        if isinstance(node, ast.Name):
            return node.id
        elif isinstance(node, ast.Attribute):
            # Handle module.Exception format
            parts = []
            current: ast.AST = node
            while isinstance(current, ast.Attribute):
                parts.append(current.attr)
                current = current.value
            if isinstance(current, ast.Name):
                parts.append(current.id)
            parts.reverse()
            return '.'.join(parts)
        return None

    def visit_Try(self, node: ast.Try) -> None:
        """Track exceptions caught in try-except blocks."""
        if self.__current_function:
            for handler in node.handlers:
                # handler.type can be None (bare except), Name, or Tuple
                if handler.type:
                    if isinstance(handler.type, ast.Name):
                        # Single exception: except ValueError:
                        exc_name = handler.type.id
                        if self.__current_function not in self.__caught_exceptions:
                            self.__caught_exceptions[self.__current_function] = set()
                        self.__caught_exceptions[self.__current_function].add(exc_name)
                    elif isinstance(handler.type, ast.Tuple):
                        # Multiple exceptions: except (ValueError, TypeError):
                        for elt in handler.type.elts:
                            caught_exc = self.__extract_exception_name(elt)
                            if caught_exc:
                                if self.__current_function not in self.__caught_exceptions:
                                    self.__caught_exceptions[self.__current_function] = set()
                                self.__caught_exceptions[self.__current_function].add(caught_exc)

        # Continue visiting child nodes
        self.generic_visit(node)

    def visit_Raise(self, node: ast.Raise) -> None:
        if self.__current_function:
            exc_name = 'Unknown'
            if isinstance(node.exc, ast.Call) and isinstance(node.exc.func, ast.Name):
                exc_name = node.exc.func.id
            elif isinstance(node.exc, ast.Name):
                exc_name = node.exc.id
            elif node.exc is None:
                # Bare raise statement
                exc_name = 'Unknown'

            if self.__current_function not in self.__explicit_raises:
                self.__explicit_raises[self.__current_function] = []
            self.__explicit_raises[self.__current_function].append((node.lineno, exc_name))

    def visit_Call(self, node: ast.Call) -> None:
        if self.__current_function:
            func_name = ''
            if isinstance(node.func, ast.Name):
                func_name = node.func.id
            # Handle simple module.func calls (e.g. os.path.join)
            elif isinstance(node.func, ast.Attribute) and isinstance(node.func.value, ast.Name):
                func_name = f'{node.func.value.id}.{node.func.attr}'
            # Handle chained attributes (e.g. os.path.join)
            elif isinstance(node.func, ast.Attribute):
                func_name = self.__resolve_attribute(node.func)

            if func_name:
                if self.__current_function not in self.__calls:
                    self.__calls[self.__current_function] = []
                self.__calls[self.__current_function].append((node.lineno, func_name))

    def __resolve_attribute(self, node: ast.Attribute) -> str:
        """Resolve chained attribute access to a string."""
        parts = []
        current: ast.AST = node
        while isinstance(current, ast.Attribute):
            parts.append(current.attr)
            current = current.value
        if isinstance(current, ast.Name):
            parts.append(current.id)
        return '.'.join(reversed(parts))



import builtins
import importlib


def get_stdlib_exceptions() -> dict[str, set[str]]:
    """Return a dictionary of standard library functions that raise exceptions."""
    return {
        # Built-in functions
        'builtins.abs': set(),
        'builtins.all': set(),
        'builtins.any': set(),
        'builtins.ascii': set(),
        'builtins.bin': set(),
        'builtins.bool': set(),
        'builtins.bytearray': set(),
        'builtins.bytes': set(),
        'builtins.callable': set(),
        'builtins.chr': {'ValueError'},
        'builtins.classmethod': set(),
        'builtins.compile': {'SyntaxError', 'ValueError'},
        'builtins.complex': {'ValueError'},
        'builtins.delattr': {'AttributeError'},
        'builtins.dict': {'ValueError'},
        'builtins.divmod': {'ZeroDivisionError'},
        'builtins.enumerate': set(),
        'builtins.eval': {'SyntaxError', 'TypeError', 'ValueError'},  # Evaluates code
        'builtins.exec': {'SyntaxError', 'TypeError', 'ValueError'},  # Executes code
        'builtins.filter': set(),
        'builtins.float': {'ValueError'},
        'builtins.format': {'ValueError'},
        'builtins.frozenset': set(),
        'builtins.getattr': {'AttributeError'},
        'builtins.globals': set(),
        'builtins.hasattr': set(),
        'builtins.hash': set(),
        'builtins.help': set(),
        'builtins.hex': set(),
        'builtins.id': set(),
        'builtins.input': {'EOFError'},  # Can raise EOFError
        'builtins.int': {'ValueError', 'TypeError'},
        'builtins.isinstance': {'TypeError'},
        'builtins.issubclass': {'TypeError'},
        'builtins.iter': {'TypeError'},
        'builtins.len': {'TypeError'},
        'builtins.list': set(),
        'builtins.locals': set(),
        'builtins.map': set(),
        'builtins.max': {'ValueError'},
        'builtins.memoryview': {'TypeError'},
        'builtins.min': {'ValueError'},
        'builtins.next': {'StopIteration'},
        'builtins.object': set(),
        'builtins.oct': set(),
        'builtins.open': {
            'FileNotFoundError',
            'PermissionError',
            'IsADirectoryError',
            'OSError',
        },
        'builtins.ord': {'TypeError'},
        'builtins.pow': {'TypeError', 'ValueError'},
        'builtins.print': {'OSError'},  # Can fail on write
        'builtins.property': set(),
        'builtins.range': {'ValueError', 'TypeError'},
        'builtins.repr': set(),
        'builtins.reversed': {'TypeError'},
        'builtins.round': {'TypeError'},
        'builtins.set': set(),
        'builtins.setattr': {'AttributeError'},
        'builtins.slice': {'TypeError'},
        'builtins.sorted': {'TypeError'},
        'builtins.staticmethod': set(),
        'builtins.str': set(),
        'builtins.sum': {'TypeError'},
        'builtins.super': {'RuntimeError'},  # If used incorrectly
        'builtins.tuple': set(),
        'builtins.type': {'TypeError'},
        'builtins.vars': {'TypeError'},
        'builtins.zip': set(),
        'builtins.__import__': {'ImportError', 'ValueError'},
        # Dict methods
        'builtins.dict.clear': set(),
        'builtins.dict.copy': set(),
        'builtins.dict.fromkeys': set(),
        'builtins.dict.get': set(),
        'builtins.dict.items': set(),
        'builtins.dict.keys': set(),
        'builtins.dict.pop': {'KeyError'},
        'builtins.dict.popitem': {'KeyError'},
        'builtins.dict.setdefault': set(),
        'builtins.dict.update': {'TypeError'},
        'builtins.dict.values': set(),
        # List methods
        'builtins.list.append': set(),
        'builtins.list.clear': set(),
        'builtins.list.copy': set(),
        'builtins.list.count': set(),
        'builtins.list.extend': set(),
        'builtins.list.index': {'ValueError'},
        'builtins.list.insert': set(),
        'builtins.list.pop': {'IndexError'},
        'builtins.list.remove': {'ValueError'},
        'builtins.list.reverse': set(),
        'builtins.list.sort': {'TypeError'},
        # Set methods
        'builtins.set.add': set(),
        'builtins.set.clear': set(),
        'builtins.set.copy': set(),
        'builtins.set.difference': set(),
        'builtins.set.difference_update': set(),
        'builtins.set.discard': set(),
        'builtins.set.intersection': set(),
        'builtins.set.intersection_update': set(),
        'builtins.set.isdisjoint': set(),
        'builtins.set.issubset': set(),
        'builtins.set.issuperset': set(),
        'builtins.set.pop': {'KeyError'},
        'builtins.set.remove': {'KeyError'},
        'builtins.set.symmetric_difference': set(),
        'builtins.set.symmetric_difference_update': set(),
        'builtins.set.union': set(),
        'builtins.set.update': set(),
        # String methods
        'builtins.str.capitalize': set(),
        'builtins.str.casefold': set(),
        'builtins.str.center': set(),
        'builtins.str.count': set(),
        'builtins.str.encode': {'UnicodeEncodeError'},
        'builtins.str.endswith': set(),
        'builtins.str.expandtabs': set(),
        'builtins.str.find': set(),
        'builtins.str.format': {'ValueError'},
        'builtins.str.format_map': {'ValueError'},
        'builtins.str.index': {'ValueError'},
        'builtins.str.isalnum': set(),
        'builtins.str.isalpha': set(),
        'builtins.str.isascii': set(),
        'builtins.str.isdecimal': set(),
        'builtins.str.isdigit': set(),
        'builtins.str.isidentifier': set(),
        'builtins.str.islower': set(),
        'builtins.str.isnumeric': set(),
        'builtins.str.isprintable': set(),
        'builtins.str.isspace': set(),
        'builtins.str.istitle': set(),
        'builtins.str.isupper': set(),
        'builtins.str.join': {'TypeError'},
        'builtins.str.ljust': set(),
        'builtins.str.lower': set(),
        'builtins.str.lstrip': set(),
        'builtins.str.maketrans': {'ValueError'},
        'builtins.str.partition': set(),
        'builtins.str.replace': set(),
        'builtins.str.rfind': set(),
        'builtins.str.rindex': {'ValueError'},
        'builtins.str.rjust': set(),
        'builtins.str.rpartition': set(),
        'builtins.str.rsplit': set(),
        'builtins.str.rstrip': set(),
        'builtins.str.split': set(),
        'builtins.str.splitlines': set(),
        'builtins.str.startswith': set(),
        'builtins.str.strip': set(),
        'builtins.str.swapcase': set(),
        'builtins.str.title': set(),
        'builtins.str.translate': set(),
        'builtins.str.upper': set(),
        'builtins.str.zfill': set(),
        # os module (file operations)
        'os.getcwd': {'OSError'},
        'os.chdir': {'FileNotFoundError', 'PermissionError', 'OSError'},
        'os.listdir': {'FileNotFoundError', 'PermissionError', 'OSError'},
        'os.mkdir': {'FileExistsError', 'FileNotFoundError', 'PermissionError', 'OSError'},
        'os.makedirs': {'FileExistsError', 'FileNotFoundError', 'PermissionError', 'OSError'},
        'os.remove': {'FileNotFoundError', 'PermissionError', 'OSError'},
        'os.rmdir': {'FileNotFoundError', 'PermissionError', 'OSError'},
        'os.rename': {'FileNotFoundError', 'PermissionError', 'OSError'},
        'os.replace': {'FileNotFoundError', 'PermissionError', 'OSError'},
        'os.link': {'FileNotFoundError', 'PermissionError', 'OSError'},
        'os.chmod': {'FileNotFoundError", "PermissionError', 'OSError'},
        'os.read': {'OSError'},
        'os.write': {'OSError'},
        # json module
        'json.dumps': {'TypeError', 'ValueError'},
        'json.loads': {'json.JSONDecodeError', 'TypeError'},
        'json.dump': {'TypeError', 'ValueError', 'OSError'},
        'json.load': {'json.JSONDecodeError', 'TypeError', 'OSError'},
    }


def _resolve_exception(name: str) -> type | None:
    """Resolve an exception name to its class."""
    # Try builtins first
    if hasattr(builtins, name):
        return getattr(builtins, name)

    # Handle dotted names (e.g. module.Error)
    if '.' in name:
        try:
            module_name, cls_name = name.rsplit('.', 1)
            module = importlib.import_module(module_name)
            return getattr(module, cls_name)
        except (ImportError, AttributeError):
            pass

    return None


def is_subtype(exc_name: str, base_name: str) -> bool:
    """
    Check if exc_name is a subtype of base_name using dynamic reflection.

    Resolves exception names to classes and uses issubclass().
    Returns False if resolution fails.
    """
    if exc_name == base_name:
        return True

    try:
        exc_cls = _resolve_exception(exc_name)
        base_cls = _resolve_exception(base_name)

        if exc_cls and base_cls:
            # Check if they are actually classes before calling issubclass
            if isinstance(exc_cls, type) and isinstance(base_cls, type):
                return issubclass(exc_cls, base_cls)
    except Exception:
        # Fail safe on any reflection error
        pass

    return False



from collections.abc import Callable
from typing import Any, TypeVar

T = TypeVar('T', bound=Callable[..., Any])


def raising(exceptions: list[type[BaseException]] | None = None) -> Callable[[T], T]:
    """
    Decorator to mark a function with the exceptions it may raise.

    Args:
        exceptions: List of exception types this function may raise.
                   Empty list [] means the function raises no exceptions.
                   None means exceptions are not tracked (default).

    Example:
        @raising(exceptions=[])  # Function raises no exceptions
        def safe_function():
            pass

        @raising(exceptions=[ValueError, TypeError])  # Function may raise these
        def risky_function(x):
            if not isinstance(x, int):
                raise TypeError
            if x < 0:
                raise ValueError

        @raising(exceptions=[FileNotFoundError, PermissionError, OSError])
        def read_config(filename):
            # Must declare exceptions from stdlib functions like open()
            with open(filename) as f:
                return f.read()
    """

    def decorator(func: T) -> T:
        return func

    return decorator



import sys


class Colors:
    """ANSI color codes for terminal output."""

    RED = '\033[91m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    BLUE = '\033[94m'
    RESET = '\033[0m'
    BOLD = '\033[1m'

    @staticmethod
    def is_tty() -> bool:
        """Check if output is to a terminal."""
        return sys.stdout.isatty()

    @classmethod
    def error(cls, text: str) -> str:
        """Format error text in red."""
        if cls.is_tty():
            return f'{cls.RED}{text}{cls.RESET}'
        return text

    @classmethod
    def success(cls, text: str) -> str:
        """Format success text in green."""
        if cls.is_tty():
            return f'{cls.GREEN}{text}{cls.RESET}'
        return text

    @classmethod
    def hint(cls, text: str) -> str:
        """Format hint text in yellow."""
        if cls.is_tty():
            return f'{cls.YELLOW}{text}{cls.RESET}'
        return text



from mypy_raise.stdlib_exceptions import is_subtype
from mypy_raise.types import (
    CallGraph,
    ExceptionPropagationMap,
    ExceptionSet,
    ExplicitRaises,
    FuncName,
    FunctionExceptionsMap,
)


def compute_exception_propagation(
    call_graph: CallGraph,
    function_exceptions: FunctionExceptionsMap,
    explicit_raises: ExplicitRaises,
    stdlib_exceptions: FunctionExceptionsMap,
    caught_exceptions: FunctionExceptionsMap,
) -> tuple[ExceptionPropagationMap, dict[FuncName, dict[FuncName, ExceptionSet]]]:
    """
    Compute which exceptions each function can raise by analyzing the call graph.

    This function traverses the call graph to determine all exceptions that can be
    raised by each function, including:
    1. Exceptions explicitly raised in the function
    2. Exceptions from called functions
    3. Exceptions from standard library functions

    Args:
        call_graph: Map of function -> list of (lineno, called_function)
        function_exceptions: Map of function -> declared exceptions
        explicit_raises: Map of function -> list of (lineno, exception)
        stdlib_exceptions: Map of stdlib function -> exceptions it raises

    Returns:
        Tuple of:
        - ExceptionPropagationMap: Map of function -> all exceptions it can raise
        - Dict mapping function -> dict of (called_func -> exceptions from that call)
    """
    # Build a map of function -> all exceptions it can actually raise
    actual_exceptions: ExceptionPropagationMap = {}

    # Track which functions contribute which exceptions (for error messages)
    exception_sources: dict[FuncName, dict[FuncName, ExceptionSet]] = {}

    # First pass: collect explicit raises for all functions
    for func_name in function_exceptions.keys():
        actual_exceptions[func_name] = set()
        exception_sources[func_name] = {}

        # Add explicitly raised exceptions
        if func_name in explicit_raises:
            for lineno, exc_name in explicit_raises[func_name]:
                actual_exceptions[func_name].add(exc_name)

    # Iteratively propagate exceptions through the call graph
    # We need multiple passes because of indirect calls
    max_iterations = 100  # Prevent infinite loops
    changed = True
    iteration = 0

    while changed and iteration < max_iterations:
        changed = False
        iteration += 1

        for func_name in function_exceptions.keys():
            if func_name not in call_graph:
                continue

            for lineno, called_func in call_graph[func_name]:
                called_exceptions: ExceptionSet = set()

                # Check if it's a decorated function
                # Check if it's a decorated function
                if called_func in function_exceptions:
                    # Get exceptions from the called function
                    if called_func in actual_exceptions:
                        called_exceptions = actual_exceptions[called_func].copy()

                # Check if it's a stdlib function (or configured custom exception)
                elif called_func in stdlib_exceptions:
                    called_exceptions = stdlib_exceptions[called_func].copy()

                # Also check with 'builtins.' prefix
                elif f'builtins.{called_func}' in stdlib_exceptions:
                    called_exceptions = stdlib_exceptions[f'builtins.{called_func}'].copy()

                # Try stripping class prefix for method calls (e.g., MyClass.method -> method)
                # Only do this if we haven't found it yet
                elif '.' in called_func:
                    method_name = called_func.split('.')[-1]
                    if method_name in function_exceptions:
                        if method_name in actual_exceptions:
                            called_exceptions = actual_exceptions[method_name].copy()

                # If we found exceptions from this call
                if called_exceptions:
                    # Track the source of these exceptions
                    if called_func not in exception_sources[func_name]:
                        exception_sources[func_name][called_func] = set()
                    exception_sources[func_name][called_func].update(called_exceptions)

                    # Add to actual exceptions
                    before_size = len(actual_exceptions[func_name])
                    actual_exceptions[func_name].update(called_exceptions)
                    if len(actual_exceptions[func_name]) > before_size:
                        changed = True

    # Subtract caught exceptions - they don't propagate
    for func_name in function_exceptions.keys():
        if func_name in caught_exceptions:
            # We must remove exceptions that are subtypes of any caught exception
            caught = caught_exceptions[func_name]

            # Filter actual_exceptions
            if func_name in actual_exceptions:
                to_remove = set()
                for exc in actual_exceptions[func_name]:
                    for caught_exc in caught:
                        if is_subtype(exc, caught_exc):
                            to_remove.add(exc)
                            break
                actual_exceptions[func_name] -= to_remove

            # Filter exception sources
            if func_name in exception_sources:
                for called_func in list(exception_sources[func_name].keys()):
                    # Identify exceptions to remove from this source
                    source_excs = exception_sources[func_name][called_func]
                    to_remove_source = set()

                    for exc in source_excs:
                        for caught_exc in caught:
                            if is_subtype(exc, caught_exc):
                                to_remove_source.add(exc)
                                break

                    exception_sources[func_name][called_func] -= to_remove_source

                    if not exception_sources[func_name][called_func]:
                        del exception_sources[func_name][called_func]

    return actual_exceptions, exception_sources


def find_undeclared_exceptions(
    function_exceptions: FunctionExceptionsMap,
    actual_exceptions: ExceptionPropagationMap,
) -> dict[FuncName, ExceptionSet]:
    """
    Find exceptions that are raised but not declared in the decorator.

    Args:
        function_exceptions: Map of function -> declared exceptions
        actual_exceptions: Map of function -> all exceptions it can raise

    Returns:
        Map of function -> undeclared exceptions
    """
    undeclared: dict[FuncName, ExceptionSet] = {}

    for func_name, declared in function_exceptions.items():
        if func_name in actual_exceptions:
            actual = actual_exceptions[func_name]
            # Instead of strict set difference, we check if each actual exception
            # is a subtype of any declared exception.
            missing = set()
            for exc in actual:
                is_handled = False
                for declared_exc in declared:
                    if is_subtype(exc, declared_exc):
                        is_handled = True
                        break
                if not is_handled:
                    missing.add(exc)

            if missing:
                undeclared[func_name] = missing

    return undeclared




import unittest
from unittest.mock import patch, MagicMock
from mypy_raise.decorator import raising
from mypy_raise.stats import AnalysisStats
from mypy_raise.colors import Colors
from mypy_raise.stdlib_exceptions import _resolve_exception, is_subtype

class TestInternals(unittest.TestCase):
    
    # helper for decorator.py
    def test_decorator_execution(self):
        """Test that the decorator actually returns the function at runtime."""
        @raising(exceptions=[ValueError])
        def func(x):
            return x + 1
        
        self.assertEqual(func(1), 2)
        
        @raising()
        def func2():
            pass
        self.assertIsNone(func2())

    # helper for stats.py
    def test_stats(self):
        stats = AnalysisStats()
        
        # Test default/empty
        self.assertEqual(stats.compliance_rate(), 100.0)
        
        # Test summary format
        summary = stats.format_summary()
        self.assertIn("Compliance rate: 100.0%", summary)
        
        # Test with data
        stats.functions_checked = 10
        stats.violations_found = 2
        self.assertEqual(stats.compliance_rate(), 80.0)
        summary = stats.format_summary()
        self.assertIn("Compliance rate: 80.0%", summary)

    # helper for stdlib_exceptions.py
    def test_resolve_exception_errors(self):
        # invalid module
        self.assertIsNone(_resolve_exception("nonexistentmodule.Error"))
        # valid module, invalid class
        self.assertIsNone(_resolve_exception("json.NonExistentError"))
    
    def test_is_subtype_errors(self):
        # Force an error during is_subtype resolution by mocking _resolve_exception
        # or passing something that breaks logic if possible.
        # However, checking the code, is_subtype handles exceptions generically.
        # Let's try to mock _resolve_exception to raise an error
        
        with patch('mypy_raise.stdlib_exceptions._resolve_exception', side_effect=ValueError("Boom")):
            self.assertFalse(is_subtype("ValueError", "Exception"))

    # helper for colors.py
    @patch('sys.stdout')
    def test_colors_tty(self, mock_stdout):
        # FORCE TTY
        mock_stdout.isatty.return_value = True
        
        self.assertIn('\033[91m', Colors.error("text"))
        self.assertIn('\033[92m', Colors.success("text"))
        self.assertIn('\033[93m', Colors.hint("text"))
        
        # FORCE NO TTY
        mock_stdout.isatty.return_value = False
        self.assertEqual(Colors.error("text"), "text")
        self.assertEqual(Colors.success("text"), "text")
        self.assertEqual(Colors.hint("text"), "text")

    # helper for visitor.py
    def test_visitor_extract_exception_name(self):
        import ast
        from mypy_raise.visitor import RaisingVisitor
        
        visitor = RaisingVisitor()
        
        # Test Name
        node = ast.Name(id='ValueError', ctx=ast.Load())
        # Access private method using name mangling
        self.assertEqual(visitor._RaisingVisitor__extract_exception_name(node), 'ValueError')
        
        # Test Attribute (e.g. os.error)
        node = ast.Attribute(
            value=ast.Name(id='os', ctx=ast.Load()),
            attr='error',
            ctx=ast.Load()
        )
        self.assertEqual(visitor._RaisingVisitor__extract_exception_name(node), 'os.error')
        
        # Test deep Attribute (e.g. a.b.c)
        node = ast.Attribute(
            value=ast.Attribute(
                value=ast.Name(id='a', ctx=ast.Load()),
                attr='b',
                ctx=ast.Load()
            ),
            attr='c',
            ctx=ast.Load()
        )
        self.assertEqual(visitor._RaisingVisitor__extract_exception_name(node), 'a.b.c')



import ast
import atexit
import configparser
import fnmatch
import sys

from mypy.nodes import MypyFile
from mypy.options import Options
from mypy.plugin import Plugin

from mypy_raise.checker import compute_exception_propagation, find_undeclared_exceptions
from mypy_raise.colors import Colors
from mypy_raise.stats import STATS
from mypy_raise.stdlib_exceptions import get_stdlib_exceptions
from mypy_raise.visitor import RaisingVisitor


def print_stats():
    """Print analysis statistics at exit."""
    if STATS.files_checked > 0:
        sys.stdout.write(STATS.format_summary() + '\n')
        sys.stdout.flush()


atexit.register(print_stats)


class RaisingPlugin(Plugin):
    def __init__(self, options: Options):
        super().__init__(options)
        self.__checked_files: set[str] = set()
        self.__stdlib_exceptions = get_stdlib_exceptions()
        self.__strict_mode = False
        self.__ignore_functions: list[str] = []
        self.__ignore_files: list[str] = []
        self.__load_config(options)

    def __load_config(self, options: Options):
        """Load configuration from mypy.ini."""
        if not options.config_file:
            return

        config = configparser.ConfigParser()
        try:
            config.read(options.config_file)
            if config.has_section('mypy-raise'):
                # Load strict mode
                self.__strict_mode = config.getboolean('mypy-raise', 'strict', fallback=False)

                # Load ignore patterns
                ignore_funcs = config.get('mypy-raise', 'ignore_functions', fallback='')
                self.__ignore_functions = [p.strip() for p in ignore_funcs.split(',') if p.strip()]

                ignore_files = config.get('mypy-raise', 'ignore_files', fallback='')
                self.__ignore_files = [p.strip() for p in ignore_files.split(',') if p.strip()]

                # Load custom exceptions (existing logic)
                self._load_custom_exceptions(config)
        except (OSError, configparser.Error):
            pass

    def _load_custom_exceptions(self, config):
        """Extract custom exception mappings from config."""
        for key, value in config['mypy-raise'].items():
            if key.startswith('exceptions_'):
                # Format: exceptions_function_name = Exception1,Exception2
                func_name = key[11:]  # Remove 'exceptions_' prefix
                exceptions = {exc.strip() for exc in value.split(',') if exc.strip()}
                self.__stdlib_exceptions[func_name] = exceptions

        # Load multiline known_exceptions
        if 'known_exceptions' in config['mypy-raise']:
            raw_data = config.get('mypy-raise', 'known_exceptions')
            # Split by lines
            lines = [line.strip() for line in raw_data.split('\n') if line.strip()]
            for line in lines:
                if ':' in line:
                    func_name, exceptions_str = line.split(':', 1)
                    func_name = func_name.strip()
                    exceptions = {exc.strip() for exc in exceptions_str.split(',') if exc.strip()}
                    self.__stdlib_exceptions[func_name] = exceptions

    def get_additional_deps(self, file: MypyFile) -> list[tuple[int, str, int]]:
        if file.fullname in self.__checked_files:
            return []

        if file.fullname.startswith(('builtins', 'typing', 'sys', 'os', 'abc', 'enum', 'mypy.', '_')):
            return []

        self.__checked_files.add(file.fullname)
        STATS.files_checked += 1

        try:
            if not file.path:
                return []

            # Check ignore patterns for file
            for pattern in self.__ignore_files:
                if fnmatch.fnmatch(file.path, pattern):
                    return []

            with open(file.path, 'r', encoding='utf-8') as f:
                source = f.read()

            tree = ast.parse(source, filename=file.path)
            visitor = RaisingVisitor()
            visitor.visit(tree)

            if not visitor.function_exceptions and not self.__strict_mode:
                # No decorated functions to check, and not in strict mode
                return []

            # Count functions analyzed
            STATS.functions_checked += len(visitor.function_exceptions)

            # Count caught exceptions
            for excs in visitor.caught_exceptions.values():
                STATS.exceptions_caught += len(excs)

            # Strict Mode: Check for undecorated functions
            if self.__strict_mode:
                self._check_strict_mode(file.path, visitor, tree)

            # Compute exception propagation
            actual_exceptions, exception_sources = compute_exception_propagation(
                call_graph=visitor.calls,
                function_exceptions=visitor.function_exceptions,
                explicit_raises=visitor.explicit_raises,
                stdlib_exceptions=self.__stdlib_exceptions,
                caught_exceptions=visitor.caught_exceptions,
            )

            # Find undeclared exceptions
            undeclared = find_undeclared_exceptions(
                function_exceptions=visitor.function_exceptions,
                actual_exceptions=actual_exceptions,
            )

            # Report errors
            for func_name, missing_exceptions in undeclared.items():
                if self._should_ignore_function(func_name):
                    continue

                lineno = visitor.function_linenos.get(func_name, 1)

                # Build detailed error message
                if missing_exceptions:
                    STATS.violations_found += 1
                    msg = self._format_error_message(
                        file_path=file.path,
                        lineno=lineno,
                        func_name=func_name,
                        missing_exceptions=missing_exceptions,
                        sources=exception_sources.get(func_name, {}),
                    )

                    sys.stdout.write(msg)
                    sys.stdout.flush()

        except Exception:
            pass

        return []

    def _check_strict_mode(self, file_path: str, visitor: RaisingVisitor, tree: ast.Module) -> None:
        """Check for functions missing @raising decorator in strict mode."""
        all_functions = self._find_all_functions(tree)
        decorated_functions = set(visitor.function_exceptions.keys())
        undecorated = all_functions - decorated_functions

        for func_name in undecorated:
            # Skip ignored functions
            if self._should_ignore_function(func_name):
                continue

            # Report at file level (line 1 approximate location)
            lineno = 1

            msg = f"{file_path}:{lineno}: {Colors.error('error')}: "
            msg += f"Function '{func_name}' missing @raising decorator (strict mode)\n"
            sys.stdout.write(msg)
            sys.stdout.flush()

    def _find_all_functions(self, tree: ast.Module) -> set[str]:
        """Find all function definitions in the file."""
        functions = set()
        for defn in tree.body:
            if isinstance(defn, (ast.FunctionDef, ast.AsyncFunctionDef)):
                functions.add(defn.name)
            elif isinstance(defn, ast.ClassDef):
                for item in defn.body:
                    if isinstance(item, (ast.FunctionDef, ast.AsyncFunctionDef)):
                        functions.add(f'{defn.name}.{item.name}')
        return functions

    def _should_ignore_function(self, func_name: str) -> bool:
        """Check if function matches ignore patterns."""
        for pattern in self.__ignore_functions:
            if fnmatch.fnmatch(func_name, pattern):
                return True
        return False

    def _format_error_message(
        self,
        file_path: str,
        lineno: int,
        func_name: str,
        missing_exceptions: set[str],
        sources: dict[str, set[str]],
    ) -> str:
        """Format enhanced error message with context and hints."""
        exc_list = ', '.join(f"'{exc}'" for exc in sorted(missing_exceptions))

        msg = f"{file_path}:{lineno}: {Colors.error('error')}: "
        msg += f"Function '{func_name}' may raise {exc_list} but these are not declared.\n"

        # Add hint
        exc_hint = ', '.join(sorted(missing_exceptions))
        hint_text = f'  💡 Hint: Add to decorator: @raising(exceptions=[{exc_hint}])'
        msg += f'{Colors.hint(hint_text)}\n'

        # Add source information
        if sources:
            msg += '  📍 Raised by:\n'
            relevant_sources = []
            for called_func, excs in sources.items():
                relevant = excs & missing_exceptions
                if relevant:
                    exc_names = ', '.join(f"'{e}'" for e in sorted(relevant))
                    relevant_sources.append(f"'{called_func}' raises {exc_names}")

            # Sort sources for deterministic output
            for source in sorted(relevant_sources):
                msg += f'     - {source}\n'

        return msg


def plugin(version: str) -> type[RaisingPlugin]:
    return RaisingPlugin
